use d3::rgb::Rgb;
use d3::vector2::Vec2Int;
pub use d3::vector3::{Direction3, Vec3};
use std::ops::Deref;

#[allow(clippy::module_name_repetitions)]
pub struct GraphicsBuffer {
    buf: Vec<Rgb>,
    /// Depth™️ buffer
    depth: Vec<u32>,
    width: usize,
    height: usize,
}

impl GraphicsBuffer {
    pub fn new() -> Self {
        Self {
            buf: Vec::new(),
            depth: Vec::new(),
            width: 0,
            height: 0,
        }
    }

    pub fn set(&mut self, pos: Vec2Int, v: Rgb, depth: u32) {
        if pos.x < 0 || pos.y < 0 || pos.x >= self.width as i32 || pos.y >= self.height as i32 {
            return;
        }
        let index = pos.x as usize + pos.y as usize * self.width;
        if depth > self.depth[index] {
            return;
        }
        self.buf[index] = v;
        // let channel = (depth / (u32::MAX / u8::MAX as u32)) as u8;
        // self.buf[index] = Rgb::new(channel, channel, channel);
        self.depth[index] = depth;
    }

    fn plot_line_low(&mut self, a: Vec2Int, b: Vec2Int, az: f32, bz: f32, v: Rgb) {
        #![allow(clippy::many_single_char_names)]
        let dx = b.x - a.x;
        let dy = (b.y - a.y).abs();
        let yi = if b.y >= a.y { 1 } else { -1 };
        let mut d = (2 * dy) - dx;
        let mut y = a.y;

        let mut z = az;
        let dz = (bz - az) / dx as f32;

        for x in a.x..b.x {
            if 0.0 < z && z < 1.0 {
                self.set((x, y).into(), v, (z * u32::MAX as f32) as u32);
            }

            if d > 0 {
                y += yi;
                d += 2 * (dy - dx);
            } else {
                d += 2 * dy;
            }

            z += dz;
        }
    }

    fn plot_line_high(&mut self, a: Vec2Int, b: Vec2Int, az: f32, bz: f32, v: Rgb) {
        #![allow(clippy::many_single_char_names)]
        let dx = (b.x - a.x).abs();
        let dy = b.y - a.y;
        let xi = if b.x >= a.x { 1 } else { -1 };
        let mut d = (2 * dx) - dy;
        let mut x = a.x;

        let mut z = az;
        let dz = (bz - az) / dy as f32;

        for y in a.y..b.y {
            if 0.0 < z && z < 1.0 {
                self.set((x, y).into(), v, (z * u32::MAX as f32) as u32);
            }

            if d > 0 {
                x += xi;
                d += 2 * (dx - dy);
            } else {
                d += 2 * dx;
            }

            z += dz;
        }
    }

    pub fn draw_line(&mut self, mut a: Vec3, mut b: Vec3, v: Rgb) {
        let quarter_width = (self.width / 4) as f32;
        let quarter_height = (self.height / 4) as f32;
        let scale = self.height as f32;

        a.x *= scale;
        b.x *= scale;
        a.y *= -scale;
        b.y *= -scale;

        a.x += quarter_width;
        b.x += quarter_width;
        a.y += quarter_height;
        b.y += quarter_height;

        // kinda hacky, need to do this to avoid the `i32` restriction in `Vec3Int`
        let az = a.z;
        let bz = b.z;

        let a = a.rounded().into_vec2();
        let b = b.rounded().into_vec2();

        // TODO: optimize for vertical and horizontal cases
        #[allow(clippy::collapsible_else_if)]
        if (b.y - a.y).abs() < (b.x - a.x).abs() {
            if a.x > b.x {
                self.plot_line_low(b, a, bz, az, v);
            } else {
                self.plot_line_low(a, b, az, bz, v);
            }
        } else {
            if a.y > b.y {
                self.plot_line_high(b, a, bz, az, v);
            } else {
                self.plot_line_high(a, b, az, bz, v);
            }
        }
    }

    pub fn resize(&mut self, dim: Vec2Int, v: Option<Rgb>) {
        let (width, height) = (dim.x as usize, dim.y as usize);
        self.buf.resize(width * height, v.unwrap_or(Rgb::BLACK));
        self.depth.resize(width * height, u32::MAX);
        self.width = width;
        self.height = height;
    }

    pub fn resize_and_clear(&mut self, dim: Vec2Int, v: Rgb) {
        if dim.x as usize != self.width || dim.y as usize != self.height {
            self.resize(dim, Some(v));
        } else {
            self.buf.fill(v);
            self.depth.fill(u32::MAX);
        }
    }
}

impl Deref for GraphicsBuffer {
    type Target = [u32];

    fn deref(&self) -> &Self::Target {
        Rgb::as_u32_slice(&self.buf)
    }
}
