use std::{
    fmt::{Debug, Display},
    ops::{Add, Div, Mul, Neg, Sub},
};

use crate::vector3::Vec3;

/// A rotation, orientation, or location in 3d space.
///
/// Recomended reading: [Quaternion on Wikipedia](https://en.wikipedia.org/wiki/Quaternion), [Quaternions and spatial rotation on Wikipedia](https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation)
///
/// Most functions and operations on `Quaternion` will require the mathematical knowledge to understand, except for [`compose`], [`apply`], [`inv`], [`rotation_around`], and [`one`].
///
/// [`compose`]: Quaternion::compose
/// [`apply`]: Quaternion::apply
/// [`inv`]: Quaternion::inv
/// [`rotation_around`]: Quaternion::rotation_around
/// [`one`]: Quaternion::one
#[derive(Copy, Clone, PartialEq, Default)]
pub struct Quaternion {
    /// The real part of the quaternion.
    pub a: f32,
    /// The coefficient of the `i` part of the quaternion.
    pub b: f32,
    /// The coefficient of the `j` part of the quaternion.
    pub c: f32,
    /// The coefficient of the `k` part of the quaternion.
    pub d: f32,
}

impl Debug for Quaternion {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{} + {}i + {}j + {}k", self.a, self.b, self.c, self.d)
    }
}

impl Display for Quaternion {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        Debug::fmt(self, f)
    }
}

///////////////////////////
// Creating a Quaternion //
///////////////////////////

impl Quaternion {
    /// Creates a scalar/real quaternion from the given value.
    pub fn scalar(a: f32) -> Self {
        Self {
            a,
            ..Default::default()
        }
    }

    /// Creates a vector quaternion from the given [`Vec3`].
    pub fn vector(v: Vec3) -> Self {
        Self {
            b: v.x,
            c: v.y,
            d: v.z,
            ..Default::default()
        }
    }

    /// Creates the zero quaternion.
    pub fn zero() -> Self {
        Quaternion::default()
    }

    /// Create a quaternion equal to 1, which is also the "no rotation" quaternion.
    pub fn one() -> Self {
        Self::scalar(1.0)
    }

    /// Creates a quaternion from the given real and vector parts.
    pub fn new(a: f32, v: Vec3) -> Self {
        Self {
            a,
            b: v.x,
            c: v.y,
            d: v.z,
        }
    }

    /// Creates a quaternion from the given components
    pub fn from_components(a: f32, b: f32, c: f32, d: f32) -> Self {
        Self { a, b, c, d }
    }

    /// Creates a quaternion representing a `theta` radians rotation around `axis`.
    pub fn rotation_around(axis: Vec3, theta: f32) -> Self {
        (theta / 2.0).cos() - Self::vector(axis.normalized()) * (theta / 2.0).sin()
    }
}

impl From<f32> for Quaternion {
    fn from(a: f32) -> Self {
        Self::scalar(a)
    }
}

impl From<Vec3> for Quaternion {
    fn from(v: Vec3) -> Self {
        Self::vector(v)
    }
}

impl From<(f32, Vec3)> for Quaternion {
    fn from((a, v): (f32, Vec3)) -> Self {
        Self::new(a, v)
    }
}

impl From<(f32, f32, f32, f32)> for Quaternion {
    fn from((a, b, c, d): (f32, f32, f32, f32)) -> Self {
        Self::from_components(a, b, c, d)
    }
}

////////////////////////////
// Operations with itself //
////////////////////////////

impl Quaternion {
    /// Gets the squared norm of the quaternion.
    pub fn sq_norm(self) -> f32 {
        self.a.powi(2) + self.b.powi(2) + self.c.powi(2) + self.d.powi(2)
    }

    /// Gets the norm of the quaternion.
    pub fn norm(self) -> f32 {
        self.sq_norm().sqrt()
    }

    /// Takes the conjugate of the quaternion.
    pub fn conjugate(self) -> Self {
        Self {
            a: self.a,
            b: -self.b,
            c: -self.c,
            d: -self.d,
        }
    }

    /// Takes the reciprocal of the quaternion.
    pub fn inv(self) -> Self {
        self.conjugate() / self.sq_norm()
    }

    /// Gets the versor/unit quaternion of the quaternion.
    pub fn versor(self) -> Self {
        self / self.norm()
    }

    /// Determines if the quaternion is a versor/unit quaternion.
    pub fn is_versor(self) -> bool {
        (self.sq_norm() - 1.0).abs() < f32::EPSILON
    }

    /// Conjugates `self` by `rhs`, the operation `rhs * self * rhs.inv()`.
    pub fn conjugate_by(self, rhs: Self) -> Self {
        rhs * self * rhs.inv()
    }

    /// Gets the scalar part of this quaternion.
    pub fn scalar_part(self) -> f32 {
        self.a
    }

    /// Gets the vector part of this quaternion.
    pub fn vector_part(self) -> Vec3 {
        (self.b, self.c, self.d).into()
    }

    /// Creates a rotation which is `self` followed by `rhs`.
    ///
    /// Same as `rhs * self`, in that order.
    pub fn compose(self, rhs: Self) -> Self {
        rhs * self
    }

    /// Composes `rhs` with `self`, in place.
    ///
    /// See also: [`Quaternion::compose`]
    pub fn compose_in_place(&mut self, rhs: Self) {
        *self = self.compose(rhs);
    }

    /// Applies this rotation to a point in space.
    ///
    /// Same as `Quaternion::vector(v).conjugate_by(self).vector_part()`
    pub fn apply(self, v: Vec3) -> Vec3 {
        Self::vector(v).conjugate_by(self).vector_part()
    }
}

impl Neg for Quaternion {
    type Output = Self;

    fn neg(self) -> Self::Output {
        Self {
            a: -self.a,
            b: -self.b,
            c: -self.c,
            d: -self.d,
        }
    }
}

impl Add for Quaternion {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        Self {
            a: self.a + rhs.a,
            b: self.b + rhs.b,
            c: self.c + rhs.c,
            d: self.d + rhs.d,
        }
    }
}

impl Sub for Quaternion {
    type Output = Self;

    fn sub(self, rhs: Self) -> Self::Output {
        self + -rhs
    }
}

impl Mul for Quaternion {
    type Output = Self;

    fn mul(self, rhs: Self) -> Self::Output {
        // https://en.wikipedia.org/wiki/Quaternion#Hamilton_product
        Self {
            a: (self.a * rhs.a) - (self.b * rhs.b) - (self.c * rhs.c) - (self.d * rhs.d),
            b: (self.a * rhs.b) + (self.b * rhs.a) + (self.c * rhs.d) - (self.d * rhs.c),
            c: (self.a * rhs.c) - (self.b * rhs.d) + (self.c * rhs.a) + (self.d * rhs.b),
            d: (self.a * rhs.d) + (self.b * rhs.c) - (self.c * rhs.b) + (self.d * rhs.a),
        }
    }
}

// `impl Div for Quaternion` is explicitly left out, since it is ambiguous.

//////////////////////////////////
// Operations with real numbers //
//////////////////////////////////

impl Add<f32> for Quaternion {
    type Output = Self;

    fn add(self, rhs: f32) -> Self::Output {
        Self {
            a: self.a + rhs,
            ..self
        }
    }
}

impl Add<Quaternion> for f32 {
    type Output = Quaternion;

    fn add(self, rhs: Quaternion) -> Self::Output {
        rhs + self
    }
}

impl Sub<f32> for Quaternion {
    type Output = Self;

    fn sub(self, rhs: f32) -> Self::Output {
        self + -rhs
    }
}

impl Sub<Quaternion> for f32 {
    type Output = Quaternion;

    fn sub(self, rhs: Quaternion) -> Self::Output {
        -rhs + self
    }
}

impl Mul<f32> for Quaternion {
    type Output = Quaternion;

    fn mul(self, rhs: f32) -> Self::Output {
        Self {
            a: self.a * rhs,
            b: self.b * rhs,
            c: self.c * rhs,
            d: self.d * rhs,
        }
    }
}

impl Mul<Quaternion> for f32 {
    type Output = Quaternion;

    fn mul(self, rhs: Quaternion) -> Self::Output {
        rhs * self
    }
}

impl Div<f32> for Quaternion {
    type Output = Quaternion;

    fn div(self, rhs: f32) -> Self::Output {
        self * (1.0 / rhs)
    }
}

// `impl Div<Quaternion> for f32` is explicitly left out, since it makes no sense
