use core::slice;

#[derive(Clone, Copy, Default)]
#[repr(transparent)]
/// An RGB (red green blue) color.
pub struct Rgb(u32);

impl Rgb {
    /// Red, or `Rgb::new(255, 0, 0)`.
    pub const RED: Self = Self::new_u32(0xff_00_00);
    /// Green, or `Rgb::new(0, 255, 0)`.
    pub const GREEN: Self = Self::new_u32(0x00_ff_00);
    /// Blue, or `Rgb::new(0, 0, 255)`.
    pub const BLUE: Self = Self::new_u32(0x00_00_ff);
    /// Yellow, or `Rgb::new(255, 255, 0)`.
    pub const YELLOW: Self = Self::new_u32(0xff_ff_00);
    /// Cyan, or `Rgb::new(0, 255, 255)`.
    pub const CYAN: Self = Self::new_u32(0x00_ff_ff);
    /// Magenta, or `Rgb::new(255, 0, 255)`.
    pub const MAGENTA: Self = Self::new_u32(0xff_00_ff);
    /// Black, or `Rgb::new(0, 0, 0)`.
    pub const BLACK: Self = Self::new_u32(0x00_00_00);
    /// White, or `Rgb::new(255, 255, 255)`.
    pub const WHITE: Self = Self::new_u32(0xff_ff_ff);

    /// Creates a new color from the given red, green, and blue values.
    pub const fn new(r: u8, g: u8, b: u8) -> Self {
        Self(((r as u32) << 16) | ((g as u32) << 8) | (b as u32))
    }

    /// Creates a new color from the given `u32`, in the following format:
    ///
    /// `00000000_RRRRRRRR_GGGGGGGG_BBBBBBBB`
    pub const fn new_u32(v: u32) -> Self {
        Self(v & 0xff_ff_ff)
    }

    /// Gets the red value of this color.
    pub fn red(self) -> u8 {
        #![allow(clippy::cast_possible_truncation)]
        (self.0 >> 16) as u8
    }

    /// Gets the green value of this color.
    pub fn green(self) -> u8 {
        ((self.0 >> 8) & 0xff) as u8
    }

    /// Gets the blue value of this color.
    pub fn blue(self) -> u8 {
        (self.0 & 0xff) as u8
    }

    /// Sets the red value of this color.
    pub fn set_red(&mut self, r: u8) {
        *self = Self((self.0 & 0x00_ff_ff) | (u32::from(r) << 16));
    }

    /// Sets the green value of this color.
    pub fn set_green(&mut self, r: u8) {
        *self = Self((self.0 & 0xff_00_ff) | (u32::from(r) << 8));
    }

    /// Sets the blue value of this color.
    pub fn set_blue(&mut self, r: u8) {
        *self = Self((self.0 & 0xff_ff_00) | u32::from(r));
    }

    /// Reinterprets an `Rgb` slice as a `u32` slice.
    pub fn as_u32_slice(slice: &[Rgb]) -> &[u32] {
        unsafe { slice::from_raw_parts((slice as *const [Rgb]).cast(), slice.len()) }
    }

    /// Takes the inner `u32` out of an `Rgb`.
    pub fn into_inner(self) -> u32 {
        self.0
    }
}

impl From<(u8, u8, u8)> for Rgb {
    fn from((r, g, b): (u8, u8, u8)) -> Self {
        Self::new(r, g, b)
    }
}

impl From<u32> for Rgb {
    fn from(v: u32) -> Self {
        Self::new_u32(v)
    }
}

impl From<Rgb> for u32 {
    fn from(v: Rgb) -> Self {
        v.into_inner()
    }
}
